/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef TX_COO_MATRIX_H
#define TX_COO_MATRIX_H

#include <cusp/coo_matrix.h>
#include <tx_common.h>
// -----------------------------------------------------------------------------

/*
	Kernel function is based on CUSP library's spmv_coo_kernel
	function which is licensed under the Apache 2.0 license.  Here
	is the license information for that part of the code.

 *  Copyright 2008-2012 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.

*/

// -----------------------------------------------------------------------------

#define CUSPMATRIXCOO cusp::coo_matrix<PetscInt,PetscScalar,cusp::device_memory>

template <typename index_type, typename value_type>
  __global__ void spmv_coo_serial_kernel(const index_type num_entries,
					 const index_type * I, 
					 const index_type * J, 
					 const value_type * V, 
					 const value_type * x, 
					 value_type * y)
{
  for(index_type n = 0; n < num_entries; n++)
    {
      y[I[n]] += V[n] * x[J[n]];
    }
}  

// segmented reduction in shared memory
template <typename index_type, typename value_type>
__device__ value_type segreduce_warp(const index_type thread_lane, index_type row, value_type val, index_type * rows, value_type * vals)
{
    rows[threadIdx.x] = row;
    vals[threadIdx.x] = val;

    if( thread_lane >=  1 && row == rows[threadIdx.x -  1] ) { vals[threadIdx.x] = val = val + vals[threadIdx.x -  1]; } 
    if( thread_lane >=  2 && row == rows[threadIdx.x -  2] ) { vals[threadIdx.x] = val = val + vals[threadIdx.x -  2]; }
    if( thread_lane >=  4 && row == rows[threadIdx.x -  4] ) { vals[threadIdx.x] = val = val + vals[threadIdx.x -  4]; }
    if( thread_lane >=  8 && row == rows[threadIdx.x -  8] ) { vals[threadIdx.x] = val = val + vals[threadIdx.x -  8]; }
    if( thread_lane >= 16 && row == rows[threadIdx.x - 16] ) { vals[threadIdx.x] = val = val + vals[threadIdx.x - 16]; }

    return val;
}

template <typename index_type, typename value_type>
__device__ void segreduce_block(const index_type * idx, value_type * val)
{
    value_type left = 0;
    if( threadIdx.x >=   1 && idx[threadIdx.x] == idx[threadIdx.x -   1] ) { left = val[threadIdx.x -   1]; } __syncthreads(); val[threadIdx.x] += left; left = 0; __syncthreads();  
    if( threadIdx.x >=   2 && idx[threadIdx.x] == idx[threadIdx.x -   2] ) { left = val[threadIdx.x -   2]; } __syncthreads(); val[threadIdx.x] += left; left = 0; __syncthreads();
    if( threadIdx.x >=   4 && idx[threadIdx.x] == idx[threadIdx.x -   4] ) { left = val[threadIdx.x -   4]; } __syncthreads(); val[threadIdx.x] += left; left = 0; __syncthreads();
    if( threadIdx.x >=   8 && idx[threadIdx.x] == idx[threadIdx.x -   8] ) { left = val[threadIdx.x -   8]; } __syncthreads(); val[threadIdx.x] += left; left = 0; __syncthreads();
    if( threadIdx.x >=  16 && idx[threadIdx.x] == idx[threadIdx.x -  16] ) { left = val[threadIdx.x -  16]; } __syncthreads(); val[threadIdx.x] += left; left = 0; __syncthreads();
    if( threadIdx.x >=  32 && idx[threadIdx.x] == idx[threadIdx.x -  32] ) { left = val[threadIdx.x -  32]; } __syncthreads(); val[threadIdx.x] += left; left = 0; __syncthreads();  
    if( threadIdx.x >=  64 && idx[threadIdx.x] == idx[threadIdx.x -  64] ) { left = val[threadIdx.x -  64]; } __syncthreads(); val[threadIdx.x] += left; left = 0; __syncthreads();
    if( threadIdx.x >= 128 && idx[threadIdx.x] == idx[threadIdx.x - 128] ) { left = val[threadIdx.x - 128]; } __syncthreads(); val[threadIdx.x] += left; left = 0; __syncthreads();
    if( threadIdx.x >= 256 && idx[threadIdx.x] == idx[threadIdx.x - 256] ) { left = val[threadIdx.x - 256]; } __syncthreads(); val[threadIdx.x] += left; left = 0; __syncthreads();
}


//////////////////////////////////////////////////////////////////////////////
// COO SpMV kernel which flattens data irregularity (segmented reduction)
//////////////////////////////////////////////////////////////////////////////
//
// spmv_coo_flat
//   The input coo_matrix must be sorted by row.  Columns within each row
//   may appear in any order and duplicate entries are also acceptable.
//   This sorted COO format is easily obtained by expanding the row pointer
//   of a CSR matrix (csr.Ap) into proper row indices and then copying 
//   the arrays containing the CSR column indices (csr.Aj) and nonzero values
//   (csr.Ax) verbatim.  A segmented reduction is used to compute the per-row
//   sums.
//
// spmv_coo_flat_tex
//   Same as spmv_coo_flat, except that the texture cache is 
//   used for accessing the x vector.
//


// spmv_coo_flat_kernel
//
// In this kernel each warp processes an interval of the nonzero values.
// For example, if the matrix contains 128 nonzero values and there are
// two warps and interval_size is 64, then the first warp (warp_id == 0)
// will process the first set of 64 values (interval [0, 64)) and the 
// second warp will process // the second set of 64 values 
// (interval [64, 128)).  Note that the  number of nonzeros is not always
// a multiple of 32 (the warp size) or 32 * the number of active warps,
// so the last active warp will not always process a "full" interval of
// interval_size.
// 
// The first thread in each warp (thread_lane == 0) has a special role:
// it is responsible for keeping track of the "carry" values from one
// iteration to the next.  The carry values consist of the row index and
// partial sum from the previous batch of 32 elements.  In the example 
// mentioned before with two warps and 128 nonzero elements, the first
// warp iterates twice and looks at the carry of the first iteration to
// decide whether to include this partial sum into the current batch.
// Specifically, if a row extends over a 32-element boundary, then the
// partial sum is carried over into the new 32-element batch.  If,
// on the other hand, the _last_ row index of the previous batch (the carry)
// differs from the _first_ row index of the current batch (the row
// read by the thread with thread_lane == 0), then the partial sum
// is written out to memory.
//
// Each warp iterates over its interval, processing 32 elements at a time.
// For each batch of 32 elements, the warp does the following
//  1) Fetch the row index, column index, and value for a matrix entry.  These
//     values are loaded from I[n], J[n], and V[n] respectively.
//     The row entry is stored in the shared memory array idx.
//  2) Fetch the corresponding entry from the input vector.  Specifically, for a 
//     nonzero entry (i,j) in the matrix, the thread must load the value x[j]
//     from memory.  We use the function fetch_x to control whether the texture
//     cache is used to load the value (UseCache == True) or whether a normal
//     global load is used (UseCache == False).
//  3) The matrix value A(i,j) (which was stored in V[n]) is multiplied by the 
//     value x[j] and stored in the shared memory array val.
//  4) The first thread in the warp (thread_lane == 0) considers the "carry"
//     row index and either includes the carried sum in its own sum, or it
//     updates the output vector (y) with the carried sum.
//  5) With row indices in the shared array idx and sums in the shared array
//     val, the warp conducts a segmented scan.  The segmented scan operation
//     looks at the row entries for each thread (stored in idx) to see whether
//     two values belong to the same segment (segments correspond to matrix rows).
//     Consider the following example which consists of 3 segments
//     (note: this example uses a warp size of 16 instead of the usual 32)
//
//           0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15   # thread_lane
//     idx [ 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2]  # row indices
//     val [ 4, 6, 5, 0, 8, 3, 2, 8, 3, 1, 4, 9, 2, 5, 2, 4]  # A(i,j) * x(j)
//
//     After the segmented scan the result will be
//
//     val [ 4,10,15,15,23,26, 2,10,13,14, 4,13,15,20,22,26]  # A(i,j) * x(j)
//
//  6) After the warp computes the segmented scan operation 
//     each thread except for the last (thread_lane == 31) looks
//     at the row index of the next thread (threadIdx.x + 1) to 
//     see if the segment ends here, or continues into the 
//     next thread.  The thread at the end of the segment writes
//     the sum into the output vector (y) at the corresponding row
//     index.
//  7) The last thread in each warp (thread_lane == 31) writes
//     its row index and partial sum into the designated spote in the
//     carry_idx and carry_val arrays.  The carry arrays are indexed
//     by warp_lane which is a number in [0, BLOCK_SIZE / 32).
//  
//  These steps are repeated until the warp reaches the end of its interval.
//  The carry values at the end of each interval are written to arrays 
//  temp_rows and temp_vals, which are processed by a second kernel.
//
template <typename index_type, typename value_type, unsigned int BLOCK_SIZE, bool UseCache>
__launch_bounds__(BLOCK_SIZE,1)
__global__ void
spmv_coo_flat_kernel(const index_type num_nonzeros,
                     const index_type interval_size,
                     const index_type * I, 
                     const index_type * J, 
                     const value_type * V, 
                     const value_type * x, 
		     value_type * y,
		     index_type * temp_rows,
		     value_type * temp_vals)
{
  __shared__ volatile index_type rows[48 *(BLOCK_SIZE/32)];
  __shared__ volatile value_type vals[BLOCK_SIZE];
  
  const index_type thread_id   = BLOCK_SIZE * blockIdx.x + threadIdx.x;                         // global thread index
  const index_type thread_lane = threadIdx.x & (WARP_SIZE-1);                                   // thread index within the warp
  const index_type warp_id     = thread_id   / WARP_SIZE;                                       // global warp index
  
  const index_type interval_begin = warp_id * interval_size;                                    // warp's offset into I,J,V
  const index_type interval_end   = thrust::min(interval_begin + interval_size, num_nonzeros);  // end of warps's work
  
  const index_type idx = 16 * (threadIdx.x/32 + 1) + threadIdx.x;                               // thread's index into padded rows array
  
  rows[idx - 16] = -1;                                                                         // fill padding with invalid row index

  if(interval_begin >= interval_end)                                                           // warp has no work to do 
    return;
  
  if (thread_lane == 31)
    {
      // initialize the carry in values
      rows[idx] = I[interval_begin]; 
      vals[threadIdx.x] = value_type(0);
    }
  
  for(index_type n = interval_begin + thread_lane; n < interval_end; n += WARP_SIZE)
    {
      index_type row = I[n];                                         // row index (i)
      value_type val = V[n] * fetch_x<UseCache>(J[n], x);            // A(i,j) * x(j)
      
      if (thread_lane == 0)
        {
	  if(row == rows[idx + 31])
                val += vals[threadIdx.x + 31];                        // row continues
	  else
	    y[rows[idx + 31]] += vals[threadIdx.x + 31];  // row terminated
        }
      
      rows[idx]         = row;
      vals[threadIdx.x] = val;
      
      if(row == rows[idx -  1]) { vals[threadIdx.x] = val = val + vals[threadIdx.x -  1]; } 
      if(row == rows[idx -  2]) { vals[threadIdx.x] = val = val + vals[threadIdx.x -  2]; }
      if(row == rows[idx -  4]) { vals[threadIdx.x] = val = val + vals[threadIdx.x -  4]; }
      if(row == rows[idx -  8]) { vals[threadIdx.x] = val = val + vals[threadIdx.x -  8]; }
      if(row == rows[idx - 16]) { vals[threadIdx.x] = val = val + vals[threadIdx.x - 16]; }
      
      if(thread_lane < 31 && row != rows[idx + 1])
	y[row] += vals[threadIdx.x];                                            // row terminated
    }
  
  if(thread_lane == 31)
    {
      // write the carry out values
      temp_rows[warp_id] = rows[idx];
      temp_vals[warp_id] = vals[threadIdx.x];
    }
}


// The second level of the segmented reduction operation
template <typename index_type, typename value_type, unsigned int BLOCK_SIZE>
__launch_bounds__(BLOCK_SIZE,1)
__global__ void
spmv_coo_reduce_update_kernel(const index_type num_warps,
                              const index_type * temp_rows,
                              const value_type * temp_vals,
                                    value_type * y)
{
  __shared__ index_type rows[BLOCK_SIZE + 1];    
  __shared__ value_type vals[BLOCK_SIZE + 1];    
  
  const index_type end = num_warps - (num_warps & (BLOCK_SIZE - 1));
  
  if (threadIdx.x == 0)
    {
      rows[BLOCK_SIZE] = (index_type) -1;
      vals[BLOCK_SIZE] = (value_type)  0;
    }
  
  __syncthreads();
  
  index_type i = threadIdx.x;
  
  while (i < end)
    {
      // do full blocks
      rows[threadIdx.x] = temp_rows[i];
      vals[threadIdx.x] = temp_vals[i];
      
      __syncthreads();
      
      segreduce_block(rows, vals);
      
      if (rows[threadIdx.x] != rows[threadIdx.x + 1])
	y[rows[threadIdx.x]] += vals[threadIdx.x];
      
      __syncthreads();
      
      i += BLOCK_SIZE; 
    }
  
  if (end < num_warps){
    if (i < num_warps){
      rows[threadIdx.x] = temp_rows[i];
      vals[threadIdx.x] = temp_vals[i];
    } else {
      rows[threadIdx.x] = (index_type) -1;
      vals[threadIdx.x] = (value_type)  0;
    }
    
    __syncthreads();
    
    segreduce_block(rows, vals);
    
    if (i < num_warps)
      if (rows[threadIdx.x] != rows[threadIdx.x + 1])
	y[rows[threadIdx.x]] += vals[threadIdx.x];
  }
}

template <typename index_type, typename value_type>
__global__ void
initialize(const index_type n,
	   const value_type initValue,
	   value_type * x)
{
  int nThreads = blockDim.x * gridDim.x;
  int tid = blockIdx.x * blockDim.x + threadIdx.x;
  for(int i = tid; i < n ; i+= nThreads)
    {
      x[i] = initValue;
    }

}

template <typename index_type, typename value_type>
  void initializeWrapper(const cudaStream_t stream,
			 const index_type n, 
			 const value_type initValue,
			 value_type *x)
{
  int nt = 128;
  int nb = (int) ceil((float) n /(float)  nt);
  int device;
  cudaError_t error = cudaGetDevice(&device);
  const int MAX_BLOCKS =  max_active_blocks(initialize<index_type, value_type>, nt, 0);
  if(nb > MAX_BLOCKS) nb = MAX_BLOCKS;
  
  dim3  grid( nb, 1, 1);
  dim3  block( nt, 1, 1);
  
  initialize<index_type, value_type><<< grid, block, 0, stream >>>( n, initValue, x);
}


template <bool UseCache, typename index_type, typename value_type>
  void spmv_coo(const cudaStream_t stream,
		const CUSPMATRIXCOO&    A, 
		const value_type* x, 
		value_type* y,
		index_type* temp_rows,
		value_type* temp_vals)
{

  const index_type * I = thrust::raw_pointer_cast(&A.row_indices[0]);
  const index_type * J = thrust::raw_pointer_cast(&A.column_indices[0]);
  const value_type * V = thrust::raw_pointer_cast(&A.values[0]);
    
  if(A.num_entries == 0)
    {
      // empty matrix
      return;
    }
  else if (A.num_entries < static_cast<size_t>(WARP_SIZE))
    {
      // small matrix
      spmv_coo_serial_kernel<index_type,value_type> <<<1,1,0,stream>>>
	(A.num_entries, I, J, V, x, y);
      return;
    }
  
  const unsigned int BLOCK_SIZE = 256;
  const unsigned int MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_coo_flat_kernel<index_type, value_type, BLOCK_SIZE, UseCache>, BLOCK_SIZE, (size_t) 0);
  const unsigned int WARPS_PER_BLOCK = BLOCK_SIZE / WARP_SIZE;
  
  const unsigned int num_units  = A.num_entries / WARP_SIZE; 
  const unsigned int num_warps  = std::min(num_units, WARPS_PER_BLOCK * MAX_BLOCKS);
  const unsigned int num_blocks = DIVIDE_INTO(num_warps, WARPS_PER_BLOCK);
  const unsigned int num_iters  = DIVIDE_INTO(num_units, num_warps);
  
  const unsigned int interval_size = WARP_SIZE * num_iters;
  
  const index_type tail = num_units * WARP_SIZE; // do the last few nonzeros separately (fewer than WARP_SIZE elements)
  
  const unsigned int active_warps = (interval_size == 0) ? 0 : DIVIDE_INTO(tail, interval_size);
  
  if (UseCache)
    bind_x(x);
  
  cusp::array1d<index_type,cusp::device_memory> tempRows(active_warps);
  cusp::array1d<value_type,cusp::device_memory> tempVals(active_warps);
  
  spmv_coo_flat_kernel<index_type, value_type, BLOCK_SIZE, UseCache> <<<num_blocks, BLOCK_SIZE, 0, stream>>>
    (tail, interval_size, I, J, V, x, y, temp_rows, temp_vals);
  //(tail, interval_size, I, J, V, x, y, thrust::raw_pointer_cast(&tempRows[0]), thrust::raw_pointer_cast(&tempVals[0]));
  
  spmv_coo_reduce_update_kernel<index_type, value_type, BLOCK_SIZE> <<<1, BLOCK_SIZE, 0, stream>>>
    (active_warps, temp_rows, temp_vals, y);
  //(active_warps, thrust::raw_pointer_cast(&tempRows[0]), thrust::raw_pointer_cast(&tempVals[0]), y);
  
  spmv_coo_serial_kernel<index_type,value_type> <<<1,1, 0, stream>>>
    (A.num_entries - tail, I + tail, J + tail, V + tail, x, y);
  
  if (UseCache)
    unbind_x(x);
}


template <bool UseCache, typename index_type, typename value_type>
  int allocate_temps(const size_t numEntries) {
  
  const unsigned int BLOCK_SIZE = 256;
  const unsigned int MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_coo_flat_kernel<index_type, value_type, BLOCK_SIZE, UseCache>, BLOCK_SIZE, (size_t) 0);
  const unsigned int WARPS_PER_BLOCK = BLOCK_SIZE / WARP_SIZE;
  
  const unsigned int num_units  = numEntries / WARP_SIZE; 
  const unsigned int num_warps  = std::min(num_units, WARPS_PER_BLOCK * MAX_BLOCKS);
  const unsigned int num_iters  = DIVIDE_INTO(num_units, num_warps);
  
  const unsigned int interval_size = WARP_SIZE * num_iters;
  
  const index_type tail = num_units * WARP_SIZE; // do the last few nonzeros separately (fewer than WARP_SIZE elements)  
  const unsigned int active_warps = (interval_size == 0) ? 0 : DIVIDE_INTO(tail, interval_size);
  return active_warps;
}

/***********************/
/*   Coo Matrix        */
/***********************/
template <typename index_type, typename value_type>
class CooMatrix {
public:
  typedef CUSPMATRIXCOO MatrixType;

  /* Constructor */
  CooMatrix() {
    temp_rows=0;
    temp_vals=0;
  }

  /* Destructor */
  virtual ~CooMatrix() {
    if (temp_rows)
      delete temp_rows;
    if (temp_vals)
      delete temp_vals;
  }

  /* Set the Matrix explicitly on the device*/
  void setMatrix(const cusp::csr_matrix<index_type,value_type,cusp::host_memory>& matCPU) {
    A = matCPU;
  }  

  /* Set the Matrix */
  void setMatrix(const index_type numRows, const index_type numCols, const index_type nz,
		 const index_type * row_offsets, const index_type * column_indices,
		 const value_type * values) {

    // create a temporary 
    cusp::csr_matrix<index_type,value_type,cusp::host_memory> B(numRows,numCols,nz);    
    B.row_offsets.assign(row_offsets, row_offsets+numRows+1);
    B.column_indices.assign(column_indices, column_indices+nz);
    B.values.assign(values, values+nz);
    // Let CUSP do the conversion
    A = B;

    temp_rows = new CUSPINTARRAYGPU;
    temp_vals = new CUSPARRAY;
    int active_warps = allocate_temps<false, index_type, value_type>(A.num_entries);

    temp_rows->resize(active_warps);
    temp_vals->resize(active_warps);

  }

  /* Set the inode data for when inode.use==true (inodes) */
  virtual void setInodeData(const index_type * inodesCPU, const int inodeCountPlusOne, const int nnzPerRowMaxCPU,
                            const int nodeMaxCPU, const int nodeCountCPU) {
    // does nothing as of now
  }


  /* cusparse SpMV */
  int cusparseMultiply(const cusparseHandle_t cusparseHandle, const cusparseMatDescr_t cusparseMatDescription,
		       const CUSPARRAY *x, CUSPARRAY *y) {
    return 0;
  }

  /* SpMV */
  int multiply(const cudaStream_t stream, const CUSPARRAY *x, CUSPARRAY *y) {
    initializeWrapper<index_type, value_type>(stream, A.num_rows, value_type(0), thrust::raw_pointer_cast(y->data()));  
    spmv_coo<false, index_type, value_type>(stream, A, 
					    thrust::raw_pointer_cast(x->data()), 
					    thrust::raw_pointer_cast(y->data()), 
					    thrust::raw_pointer_cast(temp_rows->data()), 
					    thrust::raw_pointer_cast(temp_vals->data()));
    return (int) cudaGetLastError();
  }

private:
  MatrixType A;/* container for the matrix */
  CUSPINTARRAYGPU*  temp_rows; 
  CUSPARRAY*  temp_vals; 
};

#endif // TX_COO_MATRIX_H
