/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef TX_DIA_MATRIX_H
#define TX_DIA_MATRIX_H

#include <cusp/dia_matrix.h>
#include <tx_common.h>
// -----------------------------------------------------------------------------

/*
	Kernel function is based on CUSP library's spmv_dia_kernel
	function which is licensed under the Apache 2.0 license.  Here
	is the license information for that part of the code.

 *  Copyright 2008-2012 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.

*/

// -----------------------------------------------------------------------------

////////////////////////////////////////////////////////////////////////
// DIA SpMV kernels 
///////////////////////////////////////////////////////////////////////
//
// Diagonal matrices arise in grid-based discretizations using stencils.  
// For instance, the standard 5-point discretization of the two-dimensional 
// Laplacian operator has the stencil:
//      [  0  -1   0 ]
//      [ -1   4  -1 ]
//      [  0  -1   0 ]
// and the resulting DIA format has 5 diagonals.
//
// spmv_dia
//   Each thread computes y[i] += A[i,:] * x 
//   (the dot product of the i-th row of A with the x vector)
//

#define CUSPMATRIXDIA cusp::dia_matrix<PetscInt,PetscScalar,cusp::device_memory>


template <typename index_type, typename value_type, unsigned int BLOCK_SIZE, class OPERATOR>
__launch_bounds__(BLOCK_SIZE,1)
__global__ void _spmv_dia_kernel(const index_type num_rows, 
				 const index_type num_cols, 
				 const index_type num_diagonals,
				 const index_type pitch,
				 const index_type * diagonal_offsets,
				 const value_type * values,
				 const value_type * x, 
				 value_type * y,
				 OPERATOR OP)
{
    __shared__ index_type offsets[BLOCK_SIZE];
    
    const index_type thread_id = BLOCK_SIZE * blockIdx.x + threadIdx.x;
    const index_type grid_size = BLOCK_SIZE * gridDim.x;

    for(index_type base = 0; base < num_diagonals; base += BLOCK_SIZE)
    {
        // read a chunk of the diagonal offsets into shared memory
        const index_type chunk_size = thrust::min(index_type(BLOCK_SIZE), num_diagonals - base);

        if(threadIdx.x < chunk_size)
            offsets[threadIdx.x] = diagonal_offsets[base + threadIdx.x];
    
        __syncthreads();
   
        // process chunk
        for(index_type row = thread_id; row < num_rows; row += grid_size)
        {
            value_type sum = (base == 0) ? value_type(0) : y[row];
    
            // index into values array
            index_type idx = row + pitch * base;
    
            for(index_type n = 0; n < chunk_size; n++)
            {
                const index_type col = row + offsets[n];
        
                if(col >= 0 && col < num_cols)
                {
                    const value_type A_ij = values[idx];
                    sum += A_ij * x[col];
                    //sum += A_ij * fetch_x<UseCache>(col, x);
                }
        
                idx += pitch;
            }
	    y[row] = OP(y[row],sum);
            //y[row] = sum;
        }

        // wait until all threads are done reading offsets 
        __syncthreads();
    }
}

    
template <typename index_type, typename value_type, class OPERATOR>
  void spmv_dia(const cudaStream_t stream,
		const CUSPMATRIXDIA& A,
		const value_type* x, 
		value_type* y,
		OPERATOR OP)
{
  const size_t BLOCK_SIZE = 256;
  const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(_spmv_dia_kernel<index_type, value_type, BLOCK_SIZE, OPERATOR>, BLOCK_SIZE, (size_t) sizeof(index_type) * BLOCK_SIZE);
  const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, DIVIDE_INTO(A.num_rows, BLOCK_SIZE));
  
  const index_type num_diagonals = A.values.num_cols;
  const index_type pitch         = A.values.pitch;
  
  // TODO can this be removed?
  if (num_diagonals == 0)
    {
      // empty matrix
      thrust::fill(thrust::device_pointer_cast(y), thrust::device_pointer_cast(y) + A.num_rows, value_type(0));
      return;
    }
  _spmv_dia_kernel<index_type, value_type, BLOCK_SIZE> <<<NUM_BLOCKS, BLOCK_SIZE, 0, stream>>>
    (A.num_rows, A.num_cols, num_diagonals, pitch,
     thrust::raw_pointer_cast(&A.diagonal_offsets[0]),
     thrust::raw_pointer_cast(&A.values.values[0]),
     x, y, OP);
}

/***********************/
/*   Diagonal Matrix   */
/***********************/
template <typename index_type, typename value_type>
class DiaMatrix {
public:
  typedef CUSPMATRIXDIA MatrixType;

  /* Constructor */
  DiaMatrix() {}

  /* Destructor */
  virtual ~DiaMatrix() {}

  /* Set the Matrix explicitly on the device*/
  void setMatrix(const cusp::csr_matrix<index_type,value_type,cusp::host_memory>& matCPU) {
    A = matCPU;
  }  

  /* Set the Matrix */
  void setMatrix(const index_type numRows, const index_type numCols, const index_type nz,
		 const index_type * row_offsets, const index_type * column_indices,
		 const value_type * values) {

    // create a temporary 
    cusp::csr_matrix<index_type,value_type,cusp::host_memory> B(numRows,numCols,nz);    
    B.row_offsets.assign(row_offsets, row_offsets+numRows+1);
    B.column_indices.assign(column_indices, column_indices+nz);
    B.values.assign(values, values+nz);
    // Let CUSP do the conversion
    A = B;
  }

  /* Set the Indices For cprow==true (compressed row) */
  void setCPRowIndices(const index_type * indices, const int m) {
    // does nothing as of now
  }

  /* Set the inode data for when inode.use==true (inodes) */
  virtual void setInodeData(const index_type * inodesCPU, const int inodeCountPlusOne, const int nnzPerRowMaxCPU,
                            const int nodeMaxCPU, const int nodeCountCPU) {
    // does nothing as of now
  }

  /* cusparse SpMV */
  int cusparseMultiply(const cusparseHandle_t cusparseHandle, const cusparseMatDescr_t cusparseMatDescription,
		       const CUSPARRAY *x, CUSPARRAY *y) {
    return 0;
  }

  /* SpMV with our CUSP-like kernels */
  int multiply(const cudaStream_t stream, const CUSPARRAY *x, CUSPARRAY *y) {
    spmv_dia<index_type, value_type>(stream, A, 
				     thrust::raw_pointer_cast(x->data()), 
				     thrust::raw_pointer_cast(y->data()), 
				     Equals<value_type>());
    return (int) cudaGetLastError();
  }


private:
  MatrixType A;/* container for the matrix */
};

#endif // TX_DIA_MATRIX_H
