#include <cusparse.h>
#include "tx_coo_matrix.h"
#include "tx_csr_matrix.h"
#include "tx_dia_matrix.h"
#include "tx_ell_matrix.h"
#include "tx_hyb_matrix.h"

#ifndef GPU_TIMER
#define GPU_TIMER
#endif 
#undef GPU_TIMER

//////////////////////////////////////////
// Interface class
//////////////////////////////////////////
class GPU_Matrix_Ifc {
public:

  /* Constructor */
  GPU_Matrix_Ifc() {}

  /* Destructor */
  virtual ~GPU_Matrix_Ifc() {
  }

  /* Set the Matrix */
  virtual void setMatrix(const cusp::csr_matrix<PetscInt,PetscScalar,cusp::host_memory>& matCPU) = 0;

  /* Set the Matrix */
  virtual void setMatrix(const PetscInt numRows, const PetscInt numCols, const PetscInt nz,
			 const PetscInt * row_offsets, const PetscInt * column_indices,
			 const PetscScalar * values) = 0;
  
  /* Set the Indices For cprow==true (compressed row), m = number of rows */
  virtual void setCPRowIndices(const PetscInt * indices, const int m) = 0;

  /* Set the inode data for when inode.use==true (inodes) */
  virtual void setInodeData(const PetscInt * inodes, const int inodeCountPlusOne, const int nnzPerRowMax,
                            const int nodeMax, const int nodeCount) = 0;

  /* build streams and events */
  virtual int buildStreamsAndEvents(const PetscInt MPICommSize, cudaStream_t * theBodyStream) = 0;

  /* initialize cusparse */
  virtual int initializeCusparse() = 0;

  /* needed test for synchronization */
  virtual bool hasNonZeroStream() = 0;

  /* Multiply */
  virtual int multiply(const CUSPARRAY * x, CUSPARRAY *y) = 0;

  /* Multiply Add */
  virtual int multiplyAdd(const CUSPARRAY * x, CUSPARRAY *y) = 0;
};

////////////////////////////////////////////////////
// Derived class templated over matrix type
////////////////////////////////////////////////////
template<class MTYPE>
class GPU_Matrix : public GPU_Matrix_Ifc {
public:
  
  /* Constructor */
  GPU_Matrix() {
    matClass = new MTYPE;
    indices=0;
    
    /* stream stuff */
    stream= 0;
    streamIsACopy=false;

    /* cusparse stuff */
    cusparseHandle=0;
    cusparseMatDescription=0;

    /* work vector */
    tempArray = new CUSPARRAY;

#ifdef GPU_TIMER
    startGPU=0;
    stopGPU=0;
    dt = 0.f;
    N = 0;
#endif
  }

  /* Destructor */
  virtual ~GPU_Matrix() {

    /* CUSPARSE Cleanup */
    cusparseStatus_t stat;    
    if (cusparseHandle) {
      stat = cusparseDestroy(cusparseHandle);
      if (stat!=CUSPARSE_STATUS_SUCCESS) 
	printf("Failed in ~GPU_Matrix() ... cusparseDestroy\n");
    }
    if (cusparseMatDescription) {
      stat = cusparseDestroyMatDescr(cusparseMatDescription);
      if (stat!=CUSPARSE_STATUS_SUCCESS) 
	printf("Failed in ~GPU_Matrix() ... cusparseDestroyMatDescr\n");
    }

    /* Stream Cleanup */
    if (stream && !streamIsACopy) {
      cudaError_t err = cudaStreamDestroy(stream);
      if (err!=cudaSuccess)
	printf("Failed in ~GPU_Matrix() ... cudaStreamDestroy\n");
      stream=0;
    }
#ifdef GPU_TIMER
    printf("MatMult(Add) Mean dt = %1.5f\t N=%d\n",dt/N, N);    

    // Kill the events
    if (startGPU) {
      cudaError_t err = cudaEventDestroy(startGPU);
      if (err!=cudaSuccess)
	printf("Failed in ~GPU_Matrix() ... cudaEventDestroy ... %s\n",cudaGetErrorString(err));
      startGPU=0;
    }
    if (stopGPU) {
      cudaError_t err = cudaEventDestroy(stopGPU);
      if (err!=cudaSuccess)
	printf("Failed in ~GPU_Matrix() ... cudaEventDestroy ... %s\n",cudaGetErrorString(err));
      stopGPU=0;
    }
#endif
    if (indices)
      delete indices;
    if (tempArray)
      delete tempArray;
    delete matClass;
  }

  /* Set the Matrix */
  void setMatrix(const cusp::csr_matrix<PetscInt,PetscScalar,cusp::host_memory>& matCPU) {
    matClass->setMatrix(matCPU);
  }

  /* Set the Matrix */
  void setMatrix(const PetscInt numRows, const PetscInt numCols, const PetscInt nz,
		 const PetscInt * row_offsets, const PetscInt * column_indices,
		 const PetscScalar * values) {
    
    // figure out in the class that we templated over
    matClass->setMatrix(numRows, numCols, nz, row_offsets, column_indices, values);

    /* tempArray */
    tempArray->resize(numRows);
  }

  /* Set the Indices For cprow==true (compressed row), m = number of rows */
  void setCPRowIndices(const PetscInt * cpRowIndices, const int m) {
    indices = new CUSPINTARRAYGPU;
    indices->assign(cpRowIndices,cpRowIndices+m);	
  }

  /* Set the inode data for when inode.use==true (inodes) */
  virtual void setInodeData(const PetscInt * inodes, const int inodeCountPlusOne, const int nnzPerRowMax,
                            const int nodeMax, const int nodeCount) {
    matClass->setInodeData(inodes, inodeCountPlusOne, nnzPerRowMax, nodeMax, nodeCount);
  }

  /* build streams and events */
  virtual int buildStreamsAndEvents(const PetscInt MPICommSize, cudaStream_t * theStream) {
    if (MPICommSize>1 && !(*theStream)) {
      cudaError_t err = cudaStreamCreate(&stream);
      *theStream = stream;
      if (err!=cudaSuccess) {
	printf("Failed in GPU_Matrix::buildStreamsAndEvents() ... cudaStreamCreate\n");
	return 1;
      }
    }
    else {
      // set to stream 0 if only 1 MPI rank or the input stream is 0.
      stream = *theStream;
      streamIsACopy=true;
      *theStream=0;
    }
#ifdef GPU_TIMER
    cudaError_t err = cudaEventCreate(&startGPU);
    if (err!=cudaSuccess) {
      printf("Failed in GPU_Matrix::buildStreamsAndEvents() ... cudaEventCreate ... %s\n",cudaGetErrorString(err));
      return 1;
    }
    err = cudaEventCreate(&stopGPU);
    if (err!=cudaSuccess) {
      printf("Failed in GPU_Matrix::buildStreamsAndEvents() ... cudaEventCreate ... %s\n",cudaGetErrorString(err));
      return 1;
    }
    dt=0.f;
    N=0;
#endif
    // success
    return 0;
  };

  /* initialize cusparse */
  virtual int initializeCusparse() {  
    cusparseStatus_t stat;
    stat = cusparseCreate(&cusparseHandle);
    if (stat!=CUSPARSE_STATUS_SUCCESS) {
      printf("Failed in GPU_Matrix::initializeCusparse() ... cusparseCreate ... %d\n",stat);
      return 1;
    }
    stat = cusparseCreateMatDescr(&cusparseMatDescription);
    if (stat!=CUSPARSE_STATUS_SUCCESS) {
      printf("Failed in GPU_Matrix::initializeCusparse() ... cusparseCreateMatDescr ... %d\n",stat);
      return 1;
    }
    stat = cusparseSetMatType(cusparseMatDescription,      CUSPARSE_MATRIX_TYPE_GENERAL);
    if (stat!=CUSPARSE_STATUS_SUCCESS) {
      printf("Failed in GPU_Matrix::initializeCusparse() ... cusparseSetMatType ... %d\n",stat);
      return 1;
    }
    stat = cusparseSetMatIndexBase(cusparseMatDescription, CUSPARSE_INDEX_BASE_ZERO);
    if (stat!=CUSPARSE_STATUS_SUCCESS) {
      printf("Failed in GPU_Matrix::initializeCusparse() ... cusparseSetMatIndexBase ... %d\n",stat);
      return 1;
    }
    stat = cusparseSetKernelStream(cusparseHandle, stream);
    if (stat!=CUSPARSE_STATUS_SUCCESS) {
      printf("Failed in GPU_Matrix::initializeCusparse() ... cusparseSetStream ... %d\n",stat);
      return 1;
    }
    return 0;
  }

  /* needed test for synchronization */
  virtual bool hasNonZeroStream() {
    if (stream)
      return true;
    else 
      return false;
  }

  /* Multiply */
  int multiply(const CUSPARRAY * x, CUSPARRAY *y) {

#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    int err;
    if (cusparseHandle)
      err = matClass->cusparseMultiply(cusparseHandle, cusparseMatDescription, x, y);
    else
      err = matClass->multiply(stream, x, y);

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float deltaT=0.;
    cudaEventElapsedTime(&deltaT, startGPU, stopGPU);	    
    dt+=deltaT;
    N++;
#endif
    return err;
  };

  /* Multiply Add */
  int multiplyAdd(const CUSPARRAY * x, CUSPARRAY *y) {
#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    int err;
    if (cusparseHandle)
      err = matClass->cusparseMultiply(cusparseHandle, cusparseMatDescription, x, tempArray);
    else
      err = matClass->multiply(stream, x, tempArray);

    // results are stored in a temporary array above. These are then combined in the correct places.
    thrust::for_each(thrust::make_zip_iterator(thrust::make_tuple(tempArray->begin(), thrust::make_permutation_iterator(y->begin(), indices->begin()))),
		     thrust::make_zip_iterator(thrust::make_tuple(tempArray->begin(), thrust::make_permutation_iterator(y->begin(), indices->begin()))) + tempArray->size(), 
		     ThrustPlusEquals());


#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float deltaT=0.;
    cudaEventElapsedTime(&deltaT, startGPU, stopGPU);	    
    dt+=deltaT;
    N++;
#endif
   return err;   
  };

private:
  MTYPE*  matClass;
  typename MTYPE::MatrixType mat;
  cudaStream_t      stream; /*stream for running kernels that overlap with communication */
  bool streamIsACopy;

  /* These data members are for the correction step in the SpMV add */
  CUSPARRAY*  tempArray;
  CUSPINTARRAYGPU*  indices;

  // cusparse stuff
  cusparseHandle_t cusparseHandle;
  cusparseMatDescr_t cusparseMatDescription;

#ifdef GPU_TIMER
  cudaEvent_t      startGPU;
  cudaEvent_t      stopGPU;
  float            dt;
  PetscInt         N; 
#endif

};

////////////////////////////////////////////////////
// Factory Class
////////////////////////////////////////////////////
class GPU_Matrix_Factory {
public:
  static GPU_Matrix_Ifc *getNew(std::string type){
    if (type == "coo"){
      return new GPU_Matrix<CooMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "csr"){
      return new GPU_Matrix<CsrMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "dia"){
      return new GPU_Matrix<DiaMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "ell"){
      return new GPU_Matrix<EllMatrix<PetscInt, PetscScalar> >();
    }
    else if (type == "hyb"){
      return new GPU_Matrix<HybMatrix<PetscInt, PetscScalar> >();
    }
    else {
      // default to CSR
      return new GPU_Matrix<CsrMatrix<PetscInt, PetscScalar> >();
    }
  }
};
