#include "subscript_kernels.h"

#ifndef GPU_TIMER
#define GPU_TIMER
#endif 
#undef GPU_TIMER
typedef enum {VALID_BUFFER, INVALID_BUFFER} statusFlag;

template <typename index_type, typename value_type> class GPU_Vector;

template <typename index_type, typename value_type>
class GPU_Indices {
  
 public:
  friend class GPU_Vector<index_type, value_type>;

  /////////////////////////////////////////////  
  // Constructor
  /////////////////////////////////////////////
  GPU_Indices() {
    indicesCPU = new CUSPINTARRAYCPU;
    indicesGPU = new CUSPINTARRAYGPU;
    tempGPUarray = new CUSPARRAY;
    tempCPUarray = 0;
    n = 0;
  }

  /////////////////////////////////////////////  
  // Destructor 
  /////////////////////////////////////////////
  virtual ~GPU_Indices() {
    delete indicesCPU;
    delete indicesGPU;
    delete tempGPUarray;

    // CPU work vector deallocation
    cudaError_t err = cudaFreeHost(tempCPUarray);
    if (err!=cudaSuccess)
      printf("~GPU_Indices() : Failed in cudaFreeHost ... %s\n",cudaGetErrorString(err));
  }

  /////////////////////////////////////////////  
  // buildIndices
  /////////////////////////////////////////////
  virtual int buildIndices(const index_type * indices, const index_type numElements) {

    // first sort and test contiguity
    n = numElements;
    std::vector<index_type> Vector(indices, indices+n);
    std::sort(Vector.begin(), Vector.begin()+n);
    indicesAreContiguous = true;
    for (size_t i=0; i<n-1; ++i)
      if (Vector[i+1]!=Vector[i]+1) {
	indicesAreContiguous = false;
	break;
      }
    
    //  printf("rank %d, step 1 : sendMin=%d\tsendMax=%d\tn=%d\n",rank,sendVector.front(),sendVector.back(),n);
    if (~indicesAreContiguous) {
      index_type min=0, max=0;
      min = *std::min_element(indices, indices+n);
      max = *std::max_element(indices, indices+n);
      n = max-min+1;
      //printf("rank %d, step 2 : min=%d\tmax=%d\tn=%d\n",rank,min,max,max-min+1);
      Vector.resize(n);
      for (size_t i=0; i<n; ++i)
	Vector[i] = min+i;
      indicesCPU->assign(&Vector[0],&Vector[0]+n);
      indicesGPU->assign(&Vector[0],&Vector[0]+n);
      tempGPUarray->resize(n);
      indicesAreContiguous=true;
    }
    else {
      indicesCPU->assign(indices,indices+n);
      indicesGPU->assign(indices,indices+n);
      tempGPUarray->resize(n);  
    }
    
    // CPU work vector allocation
    cudaError_t err = cudaMallocHost((void **) &tempCPUarray, n*sizeof(value_type));
    if (err!=cudaSuccess) {
      printf("GPU_Indices::buildIndices() : Failed in cudaMallocHost ... %s\n",cudaGetErrorString(err));
      return 1;
    }
    valid_GPU_array = INVALID_BUFFER;
    return 0;
  }
  
  /////////////////////////////////////////////  
  // resetFlag
  /////////////////////////////////////////////
  virtual void resetStatusFlag() {
    valid_GPU_array = INVALID_BUFFER;
  }

  /////////////////////////////////////////////  
  // acccesors
  /////////////////////////////////////////////
  
 private:
  
  /* CPU indices */
  CUSPINTARRAYCPU * indicesCPU;
  /* GPU indices */
  CUSPINTARRAYGPU * indicesGPU;

  /* number of indices */
  index_type n;

  /* this is a work vector for copying noncontiguous data to continguous buffers and vice versa */
  CUSPARRAY * tempGPUarray;

  /* this is a CPU work vector for copying noncontiguous data to continguous buffers and vice versa */
  value_type * tempCPUarray; 

  /* indicates where the most recently modified vector data is */
  statusFlag valid_GPU_array;    

  /* indicates whether indices are contiguous or not */
  bool indicesAreContiguous;
};

template <typename index_type, typename value_type>
class GPU_Vector {
 public:

  /////////////////////////////////////////////  
  // Constructor
  /////////////////////////////////////////////
  GPU_Vector(CUSPARRAY * gpuVector, size_t theRank) {
    /* arrays, CPU and GPU */

    // This will be allocated subsequently in another
    // function reallocateHostMemory via cudaMallocHost
    CPUvector = 0;

    // At the moment the GPU vector created/allocated 
    // somewhere else. Thus, we do not delete it
    // in the destructor
    GPUvector = gpuVector;
    n = gpuVector->size();

    /* stream stuff */
    stream= 0;
    ownsPtr = false;

    rank=theRank;

#ifdef GPU_TIMER
    /* timer stuff */
    startGPU=0;
    stopGPU=0;
    
    dtSubscriptLHS=0.f;
    dtSubscriptRHS=0.f;
    dtFrom=0.f;
    dtTo=0.f;
    NSubscriptLHS=0;
    NSubscriptRHS=0;
    NFrom=0;
    NTo=0;
#endif
  }

  /////////////////////////////////////////////  
  // Destructor 
  /////////////////////////////////////////////
  virtual ~GPU_Vector() {
    
    // Kill the stream
    cudaError_t err = cudaStreamDestroy(stream);
    if (err!=cudaSuccess)
      printf("~GPU_Vector() : Failed in cudaStreamDestroy ... %s\n",cudaGetErrorString(err));
    stream=0;

    // delete the CPU ptr, if owned.
    if (ownsPtr) {
      err = cudaFreeHost(CPUvector); 
      if (err!=cudaSuccess)
        printf("~GPU_Vector() : Failed in cudaFreeHost on %p ... %s\n",CPUvector,cudaGetErrorString(err));
    }

#ifdef GPU_TIMER
    printf("Mean dtFrom = %1.5f\t NFrom=%d\n",dtFrom/NFrom,NFrom);
    printf("Mean dtTo   = %1.5f\t NTo=%d\n",dtTo/NTo,NTo);
    printf("Mean dtSubscriptRHS   = %1.5f\t NSubscriptRHS=%d\n",dtSubscriptRHS/NSubscriptRHS,NSubscriptRHS);
    printf("Mean dtSubscriptLHS   = %1.5f\t NSubscriptLHS=%d\n",dtSubscriptLHS/NSubscriptLHS,NSubscriptLHS);
    
    // Kill the events
    if (startGPU) {
      err = cudaEventDestroy(startGPU);
      if (err!=cudaSuccess)
	printf("~GPU_Vector() : Failed in cudaEventDestroy ... %s\n",cudaGetErrorString(err));
      startGPU=0;
    }
    if (stopGPU) {
      err = cudaEventDestroy(stopGPU);
      if (err!=cudaSuccess)
	printf("~GPU_Vector() : Failed in cudaEventDestroy ... %s\n",cudaGetErrorString(err));
      stopGPU=0;
    }
    
    // reset vars
    dtSubscriptRHS=0.f;
    dtSubscriptLHS=0.f;
    dtFrom=0.f;
    dtTo=0.f;
    NSubscriptRHS=0;
    NSubscriptLHS=0;
    NFrom=0;
    NTo=0;
#endif
  }


  /////////////////////////////////////////////
  // allocate host memory as page locked 
  /////////////////////////////////////////////
  virtual int allocateHostMemory() {
    cudaError_t err = cudaMallocHost((void **) &CPUvector, n*sizeof(value_type));
    ownsPtr=true;
    return (int) err;
  }
  virtual value_type * getHostMemoryPtr() {
    return CPUvector;
  }

  virtual void assignHostMemory(value_type * hostVector) {
    ownsPtr=false;
    CPUvector = hostVector;
  }
  
  /* build streams and events */
  virtual int buildStreamsAndEvents() {
    
    cudaError_t err;
    err = cudaStreamCreate(&stream);
    if (err!=cudaSuccess) {
      printf("GPU_Vector::buildStreamsAndEvents() : Failed in cudaStreamCreate ... %s\n",cudaGetErrorString(err));
      return 1;
    }

#ifdef GPU_TIMER
    err = cudaEventCreate(&startGPU);
    if (err!=cudaSuccess) {
      printf("GPU_Vector::buildStreamsAndEvents() : Failed in cudaEventCreate ... %s\n",cudaGetErrorString(err));
      return 1;
    }
    err = cudaEventCreate(&stopGPU);
    if (err!=cudaSuccess) {
      printf("GPU_Vector::buildStreamsAndEvents() : Failed in cudaEventCreate ... %s\n",cudaGetErrorString(err));
      return 1;
    }
#endif
    return 0;
  };

  //////////////////////////////
  // CopyToGPUAll
  //////////////////////////////
  virtual int copyToGPUAll() {

#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif    
    value_type * gpuPtr = thrust::raw_pointer_cast(&(GPUvector->begin())[0]);
    //printf("%d : copyToGPUAll for %p\n",rank,CPUvector);
    //for(size_t i=0; i<n; ++i)
    // printf("%d : %1.5g\n",rank,CPUvector[i]);

    cudaError_t err = cudaMemcpyAsync(gpuPtr, 
				      CPUvector, n*sizeof(value_type), 
				      cudaMemcpyHostToDevice, stream);
    if (err!=cudaSuccess) {
      printf("GPU_Vector::copyToGPUAll() : Failed in cudaMemcpyAsync ... %s\n",cudaGetErrorString(err));
      return 1;
    }

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtTo+=dt;
    NTo++;
#endif

    //Have to synchronize the stream 
    err = cudaStreamSynchronize(stream);
    if (err!=cudaSuccess) {
      printf("GPU_Vector::copyToGPUAll() : Failed in cudaStreamSynchronize ... %s\n",cudaGetErrorString(err));
      return 1;
    }
    return 0;    
  }



  //////////////////////////////
  // CopyToGPUSome
  //////////////////////////////
  virtual int copyToGPUSome(CUSPARRAY* varray,
			    GPU_Indices<index_type, value_type>* indices) {

    value_type *tempCPUarray=indices->tempCPUarray;
    CUSPARRAY *tempGPUarray=indices->tempGPUarray;
    CUSPINTARRAYGPU *indicesGPU=indices->indicesGPU;
    CUSPINTARRAYCPU *indicesCPU=indices->indicesCPU;
    index_type m = indicesGPU->size();

    // Going to do this in 3 steps
    // 1) Move Scattered array to contiguous array on the host
    // 2) Move contiguous array to device via cudaMemcpyAsync()
    // 3) Move contiguous device data to scattered data on the device.

    // Step 1: copy into correct locations
    index_type * cpuInd = thrust::raw_pointer_cast(&(indicesCPU->begin())[0]);
    for (size_t i=0; i<m; ++i)
      tempCPUarray[i] = CPUvector[cpuInd[i]];
    //printf("%d : copyToGPUSome for %p\n",rank,CPUvector);
    //for(size_t i=0; i<m; ++i)
    // printf("%d : (%d, %1.5g)\n",rank,tempCPUarray[i],cpuInd[i]);



    // Step 2: asynchronous memory transfer to the device
#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    value_type * gpuPtr;
    if (!indices->indicesAreContiguous)
      gpuPtr = thrust::raw_pointer_cast(&(tempGPUarray->begin())[0]);
    else 
      gpuPtr = thrust::raw_pointer_cast(&(varray->begin())[0])+cpuInd[0];
        
    cudaError_t err = cudaMemcpyAsync(gpuPtr, tempCPUarray, m*sizeof(value_type), 
				      cudaMemcpyHostToDevice, stream); 
    if (err!=cudaSuccess) {
      printf("GPU_Vector::copyToGPUSome() : Failed in cudaMemcpyAsync ... %s\n",cudaGetErrorString(err));
      return 1;
    }

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtTo+=dt;
    NTo++;
#endif
    indices->valid_GPU_array = VALID_BUFFER;

    // Step 3: this function scatters data from the contiguous buffer to the correct locations in the vector
    // on the device.
    if (!indices->indicesAreContiguous) {
      int err = copySomeFromContiguousBuffer(varray, indices);
      if (err>0) return 1;
    }

    //Have to synchronize the stream 
    //err = cudaThreadSynchronize();
    err = cudaStreamSynchronize(stream);
    if (err!=cudaSuccess) {
      printf("GPU_Vector::copyToGPUSome() : Failed in cudaStreamSynchronize ... %s\n",cudaGetErrorString(err));
      return 1;
    }    
    return 0;
  }

  //////////////////////////////
  // CopyFromGPUAll
  //////////////////////////////
  virtual int copyFromGPUAll() {

#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    value_type * gpuPtr = thrust::raw_pointer_cast(&(GPUvector->begin())[0]);
    cudaError_t err = cudaMemcpyAsync(CPUvector, gpuPtr, 
				      n*sizeof(value_type), cudaMemcpyDeviceToHost, stream);
    if (err!=cudaSuccess) {
      printf("GPU_Vector::copyFromGPUAll() : Failed in cudaMemcpyAsync ... %s\n",cudaGetErrorString(err));
      return 1;
    }
    
    //Have to synchronize the stream before the MPI buffers are filled.
    err = cudaStreamSynchronize(stream);
    if (err!=cudaSuccess) {
      printf("GPU_Vector::copyFromGPUAll() : Failed in cudaStreamSynchronize ... %s\n",cudaGetErrorString(err));
      return 1;
    }

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtFrom+=dt;
    NFrom++;      
#endif
    return 0;    
  }
  
  
  //////////////////////////////
  // copyFromGPUSome
  //////////////////////////////
  virtual int copyFromGPUSome(CUSPARRAY* varray,
			      GPU_Indices<index_type, value_type>* indices) {

    value_type *tempCPUarray=indices->tempCPUarray;
    CUSPARRAY *tempGPUarray=indices->tempGPUarray;
    CUSPINTARRAYGPU *indicesGPU=indices->indicesGPU;
    CUSPINTARRAYCPU *indicesCPU=indices->indicesCPU;
    index_type m = indicesGPU->size();

    // Going to do this in 3 steps
    // 1) Move Scattered array to contiguous array on the device ... this is done in VecCUSPCopySomeToContiguousBufferGPU
    // 2) Move contiguous array to host via cudaMemcpyAsync()
    // 3) On the CPU, move data in contiguous buffer to the correct scattered locations.

    // Step 1:
    if (!indices->indicesAreContiguous) {
      int err = copySomeToContiguousBuffer(varray, indices);
      if (err>0) return 1;
    }

      // Step 2:
#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif
    value_type * gpuPtr;
    index_type * cpuInd = thrust::raw_pointer_cast(&(indicesCPU->begin())[0]);
    if (!indices->indicesAreContiguous)
      gpuPtr = thrust::raw_pointer_cast(&(tempGPUarray->begin())[0]);
    else
      gpuPtr = thrust::raw_pointer_cast(&(varray->begin())[0])+cpuInd[0];

    cudaError_t err = cudaMemcpyAsync(tempCPUarray, gpuPtr, m*sizeof(value_type), 
				      cudaMemcpyDeviceToHost, stream); 
    if (err!=cudaSuccess) {
      printf("GPU_Vector::copyFromGPUSome() : Failed in cudaMemcpyAsync ... %s\n",cudaGetErrorString(err));
      return 1;
    }
    
    //Have to synchronize the stream 
    err = cudaStreamSynchronize(stream);
    if (err!=cudaSuccess) {
      printf("GPU_Vector::copyFromGPUSome() : Failed in cudaStreamSynchronize ... %s\n",cudaGetErrorString(err));
      return 1;
    }

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtFrom+=dt;
    NFrom++;
#endif

    //printf("%d : copyFromGPUSome for %p\n",rank, CPUvector);
    //for(size_t i=0; i<m; ++i)
    // printf("%d : (%d, %1.5g)\n",rank,tempCPUarray[i],cpuInd[i]);

    // Step 3: now copy into correct locations
    for (size_t i=0; i<m; ++i)
      CPUvector[cpuInd[i]] = tempCPUarray[i];
    return 0;
  }


  //////////////////////////////
  // copySomeToContiguousBuffer
  //////////////////////////////
  virtual int copySomeToContiguousBuffer(CUSPARRAY* varray,
					 GPU_Indices<index_type, value_type>* indices) {

    CUSPARRAY *tempGPUarray=indices->tempGPUarray;
    CUSPINTARRAYGPU *indicesGPU=indices->indicesGPU;
    index_type m = indicesGPU->size();

#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif

    if (indices->valid_GPU_array==INVALID_BUFFER && !indices->indicesAreContiguous) {
      index_type * indPtr = thrust::raw_pointer_cast(&(indicesGPU->begin())[0]);
      value_type * gpuPtrLHS = thrust::raw_pointer_cast(&(varray->begin())[0]);
      value_type * gpuPtrRHS = thrust::raw_pointer_cast(&(tempGPUarray->begin())[0]);

      cudaError_t err = subscriptRHS(stream, m, indPtr, gpuPtrLHS, gpuPtrRHS);

      if (err!=cudaSuccess) {
	printf("GPU_Vector::copySomeToContiguousBuffer() : Failed in subscriptRHS ... %s\n",cudaGetErrorString(err));
	return 1;
      }
    }
    indices->valid_GPU_array=VALID_BUFFER;

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);	    
    dtSubscriptRHS+=dt;
    NSubscriptRHS++;
#endif
    return 0;
  }


  ////////////////////////////////////
  // copySomeFromContiguousBuffer
  ////////////////////////////////////
  virtual int copySomeFromContiguousBuffer(CUSPARRAY* varray,
					   GPU_Indices<index_type, value_type>* indices) {

    CUSPARRAY *tempGPUarray=indices->tempGPUarray;
    CUSPINTARRAYGPU *indicesGPU=indices->indicesGPU;
    index_type m = indicesGPU->size();

#ifdef GPU_TIMER
    cudaEventRecord(startGPU, stream);
#endif

    if (indices->valid_GPU_array==VALID_BUFFER) {
      index_type * indPtr = thrust::raw_pointer_cast(&(indicesGPU->begin())[0]);
      value_type * gpuPtrRHS = thrust::raw_pointer_cast(&(tempGPUarray->begin())[0]);
      value_type * gpuPtrLHS = thrust::raw_pointer_cast(&(varray->begin())[0]);
      
      cudaError_t err = subscriptLHS(stream, m, indPtr, gpuPtrRHS, gpuPtrLHS);

      if (err!=cudaSuccess) {
	printf("GPU_Vector::copySomeFromContiguousBuffer() : Failed in subscriptLHS ... %s\n",cudaGetErrorString(err));
	return 1;
      }
    }

#ifdef GPU_TIMER
    cudaEventRecord(stopGPU, stream);
    cudaEventSynchronize(stopGPU);
    float dt=0.;
    cudaEventElapsedTime(&dt, startGPU, stopGPU);
    dtSubscriptLHS+=dt;
    NSubscriptLHS++;
#endif
    return 0;
  }

private:

  cudaStream_t     stream;    /* it's easier to put it in here */
  value_type *   CPUvector;  /* pointer to page-locked host memory */
  CUSPARRAY*     GPUvector;  /* this always holds the ptr to GPU data */
  index_type     n; /* size of the vector */
  bool           ownsPtr;
  size_t         rank;

#ifdef GPU_TIMER
  cudaEvent_t      startGPU;
  cudaEvent_t      stopGPU;
  float      dtSubscriptLHS;
  float      dtSubscriptRHS;
  float      dtFrom;
  float      dtTo;
  index_type NSubscriptLHS; /* number of times that VecCUSPCopyFromGPU has been called */
  index_type NSubscriptRHS; /* number of times that VecCUSPCopyFromGPU has been called */
  index_type NFrom; /* number of times that VecCUSPCopyFromGPU has been called */
  index_type NTo; /* number of times that VecCUSPCopyToGPU has been called */
#endif
};
